"""
loop_bkg.py corresponding to draft/loop_bkg_cell, which calculates E and T from cell imputed matrix
merge_cell_to_group.py corresponding to loop_sumcell_chr,
which aggregates the cell E, T to cluster level E, E2, T, T2, O,
"""

from .snakemake import call_loop
